@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2015 DNAnexus, Inc.
+# Copyright (C) 2013-2016 DNAnexus, Inc.
 #
 # This file is part of dx-toolkit (DNAnexus platform client libraries).
 #
@@ -125,20 +125,49 @@ environment variables:
 
 from __future__ import print_function, unicode_literals, division, absolute_import
 
-import os, sys, json, time, logging, platform, ssl, traceback
-import errno
-import requests
-import socket
-from collections import namedtuple
-from . import exceptions
-from requests.auth import AuthBase
-from requests.packages import urllib3
-from requests.packages.urllib3.packages.ssl_match_hostname import match_hostname
-from .compat import USING_PYTHON2, expanduser
+import logging
 
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.NullHandler())
 
+import os, sys, json, time, platform, ssl, traceback
+import errno
+import math
+import mmap
+import requests
+import socket
+import threading
+import subprocess
+
+from collections import namedtuple
+
+from . import exceptions
+from .compat import USING_PYTHON2, BadStatusLine, StringIO, bytes, Repr
+from .utils.printing import BOLD, BLUE, YELLOW, GREEN, RED, WHITE
+
+from random import randint
+from requests.auth import AuthBase
+from requests.packages import urllib3
+from requests.packages.urllib3.packages.ssl_match_hostname import match_hostname
+from threading import Lock
+
+try:
+    from urllib.parse import urlsplit
+except ImportError:
+    from urlparse import urlsplit
+
+
+sequence_number_mutex = threading.Lock()
+counter = 0
+
+
+def _get_sequence_number():
+    global counter
+    with sequence_number_mutex:
+        retval = counter
+        counter += 1
+        return retval
+
 
 def configure_urllib3():
     # Disable verbose urllib3 warnings and log messages
@@ -172,39 +201,92 @@ APISERVER_PORT = DEFAULT_APISERVER_PORT
 
 DEFAULT_RETRIES = 6
 DEFAULT_TIMEOUT = 600
-DEFAULT_RETRY_AFTER_503_INTERVAL = 60
 
 _DEBUG = 0  # debug verbosity level
 _UPGRADE_NOTIFY = True
 
+INCOMPLETE_READS_NUM_SUBCHUNKS = 8
+
 USER_AGENT = "{name}/{version} ({platform})".format(name=__name__,
                                                     version=TOOLKIT_VERSION,
                                                     platform=platform.platform())
+_default_certs = requests.certs.where()
 _default_headers = requests.utils.default_headers()
-_default_headers['DNAnexus-API'] = API_VERSION
-_default_headers['User-Agent'] = USER_AGENT
 _default_timeout = urllib3.util.timeout.Timeout(connect=DEFAULT_TIMEOUT, read=DEFAULT_TIMEOUT)
-_pool_manager = urllib3.PoolManager(maxsize=32,
-                                    cert_reqs=ssl.CERT_REQUIRED,
-                                    ca_certs=requests.certs.where(),
-                                    headers=_default_headers,
-                                    timeout=_default_timeout)
 _RequestForAuth = namedtuple('_RequestForAuth', 'method url headers')
-_expected_exceptions = exceptions.network_exceptions + (exceptions.DXAPIError, )
+_expected_exceptions = (exceptions.network_exceptions, exceptions.DXAPIError, BadStatusLine, exceptions.BadJSONInReply,
+                        exceptions.UrllibInternalError)
 
-def _get_pool_manager(request_kwargs):
-    if 'verify' in request_kwargs or 'DX_CA_CERT' in os.environ:
-        cert_reqs = ssl.CERT_REQUIRED
-        ca_certs = request_kwargs.get('verify', os.environ.get('DX_CA_CERT'))
-        if request_kwargs.get('verify') is False or os.environ.get('DX_CA_CERT') == 'NOVERIFY':
-            cert_reqs, ca_certs = ssl.CERT_NONE, None
-            urllib3.disable_warnings()
-        return urllib3.PoolManager(cert_reqs=cert_reqs,
-                                   ca_certs=ca_certs,
-                                   headers=_default_headers,
-                                   timeout=_default_timeout)
+# Multiple threads can ask for the pool, so we need to protect
+# access and make it thread safe.
+_pool_mutex = Lock()
+_pool_manager = None
+
+def _get_proxy_info(url):
+    proxy_info = {}
+
+    url_info = urlsplit(url)
+    # If the url contains a username, need to separate the username/password
+    # from the url
+    if url_info.username:
+        # Strip the username/password out of the url
+        url = url_info.netloc[url_info.netloc.find('@')+1:]
+        # Now get the username and possibly password
+        proxy_info['proxy_url'] = '{0}://{1}'.format(url_info.scheme, url)
+        if url_info.password:
+            proxy_auth = '{0}:{1}'.format(url_info.username, url_info.password)
+        else:
+            proxy_auth = url_info.username
+        proxy_info['proxy_headers'] = urllib3.make_headers(proxy_basic_auth=proxy_auth)
     else:
-        return _pool_manager
+        # No username was given, so just take the url as is.
+        proxy_info['proxy_url'] = url
+
+    return proxy_info
+
+def _get_env_var_proxy(print_proxy=False):
+  proxy_tuple = ('http_proxy', 'HTTP_PROXY', 'https_proxy', 'HTTPS_PROXY')
+  proxy = None
+  for env_proxy in proxy_tuple:
+    if env_proxy in os.environ:
+      proxy = os.environ[env_proxy]
+  if print_proxy:
+    print('Using env variable %s=%s as proxy' % (env_proxy,proxy))
+  return proxy
+
+def _get_pool_manager(verify, cert_file, key_file):
+    global _pool_manager
+    default_pool_args = dict(maxsize=32,
+                             cert_reqs=ssl.CERT_REQUIRED,
+                             ca_certs=_default_certs,
+                             headers=_default_headers,
+                             timeout=_default_timeout)
+    if cert_file is None and verify is None and 'DX_CA_CERT' not in os.environ:
+        with _pool_mutex:
+            if _pool_manager is None:
+                if _get_env_var_proxy():
+                    proxy_params = _get_proxy_info(_get_env_var_proxy(print_proxy=True))
+                    default_pool_args.update(proxy_params)
+                    _pool_manager = urllib3.ProxyManager(**default_pool_args)
+                else:
+                    _pool_manager = urllib3.PoolManager(**default_pool_args)
+            return _pool_manager
+    else:
+        # This is the uncommon case, normally, we want to cache the pool
+        # manager.
+        pool_args = dict(default_pool_args,
+                         cert_file=cert_file,
+                         key_file=key_file,
+                         ca_certs=verify or os.environ.get('DX_CA_CERT') or requests.certs.where())
+        if verify is False or os.environ.get('DX_CA_CERT') == 'NOVERIFY':
+            pool_args.update(cert_reqs=ssl.CERT_NONE, ca_certs=None)
+            urllib3.disable_warnings()
+        if _get_env_var_proxy():
+            proxy_params = _get_proxy_info(_get_env_var_proxy(print_proxy=True))
+            pool_args.update(proxy_params)
+            return urllib3.ProxyManager(**pool_args)
+        else:
+            return urllib3.PoolManager(**pool_args)
 
 
 def _process_method_url_headers(method, url, headers):
@@ -266,25 +348,139 @@ def _extract_msg_from_last_exception():
         return traceback.format_exc().splitlines()[-1].strip()
 
 
-def _extract_retry_after_timeout(response):
-    '''Returns the time in seconds that the server is asking us to
-    wait. The information is deduced from the server http response.'''
+def _calculate_retry_delay(response, num_attempts):
+    '''
+    Returns the time in seconds that we should wait.
+
+    :param num_attempts: number of attempts that have been made to the
+        resource, including the most recent failed one
+    :type num_attempts: int
+    '''
+    if response is not None and response.status == 503 and 'retry-after' in response.headers:
+        try:
+            return int(response.headers['retry-after'])
+        except ValueError:
+            # In RFC 2616, retry-after can be formatted as absolute time
+            # instead of seconds to wait. We don't bother to parse that,
+            # but the apiserver doesn't generate such responses anyway.
+            pass
+    if num_attempts <= 1:
+        return 1
+    num_attempts = min(num_attempts, 7)
+    return randint(2 ** (num_attempts - 2), 2 ** (num_attempts - 1))
+
+
+# Truncate the message, if the error injection flag is on, and other
+# conditions hold. This causes a BadRequest 400 HTTP code, which is
+# subsequentally retried.
+#
+# Note: the minimal upload size for S3 is 5MB. In theory, you are
+# supposed to get an "EntityTooSmall" error from S3, which has a 400
+# code. However, I have not observed such responses in practice.
+# http://docs.aws.amazon.com/AmazonS3/latest/API/ErrorResponses.html
+def _maybe_truncate_request(url, data):
+    MIN_UPLOAD_LEN = 16 * 1024
+    if _INJECT_ERROR:
+        if (randint(0, 9) == 0) and "upload" in url and len(data) > MIN_UPLOAD_LEN:
+            logger.info("truncating upload data to length=%d", MIN_UPLOAD_LEN)
+            return data[0:MIN_UPLOAD_LEN]
+    return data
+
+
+def _raise_error_for_testing(try_index=None, method='GET'):
+    if _INJECT_ERROR and method == 'GET' and randint(0, 9) == 0:
+        error_thrown = randint(0, 1)
+        if error_thrown == 0 and try_index is None:
+            raise exceptions.DXIncompleteReadsError()
+
+        # Raise exception to test urllib3 error in downloads
+        elif error_thrown == 1 and try_index is not None and try_index < 3:
+            raise exceptions.UrllibInternalError()
+
+
+def _debug_print_request(debug_level, seq_num, time_started, method, url, headers, jsonify_data, data):
+    if debug_level >= 2:
+        if not jsonify_data:
+            if len(data) == 0:
+                formatted_data = '""'
+            else:
+                formatted_data = "<file data of length " + str(len(data)) + ">"
+        else:
+            try:
+                if _DEBUG >= 3:
+                    formatted_data = json.dumps(data, indent=2)
+                else:
+                    formatted_data = json.dumps(data)
+            except (UnicodeDecodeError, TypeError):
+                formatted_data = "<binary data>"
+
+        printable_headers = ''
+        if 'Range' in headers:
+            printable_headers = " " + json.dumps({"Range": headers["Range"]})
+        print("%s [%f] %s %s%s => %s\n" % (YELLOW(BOLD(">%d" % seq_num)),
+                                           time_started,
+                                           BLUE(method),
+                                           url,
+                                           printable_headers,
+                                           formatted_data),
+              file=sys.stderr,
+              end="")
+    elif debug_level > 0:
+        print("%s [%f] %s %s => %s\n" % (YELLOW(BOLD(">%d" % seq_num)),
+                                         time_started,
+                                         BLUE(method),
+                                         url,
+                                         Repr().repr(data)),
+              file=sys.stderr,
+              end="")
+
+
+def _debug_print_response(debug_level, seq_num, time_started, req_id, response_status, response_was_json, method,
+                          url, content):
+    if debug_level > 0:
+        if response_was_json:
+            if debug_level >= 3:
+                content_to_print = "\n  " + json.dumps(content, indent=2).replace("\n", "\n  ")
+            elif debug_level == 2:
+                content_to_print = json.dumps(content)
+            else:
+                content_to_print = Repr().repr(content)
+        else:
+            content_to_print = "(%d bytes)" % len(content) if len(content) > 0 else ''
+
+        t = int((time.time() - time_started) * 1000)
+        code_format = GREEN if (200 <= response_status < 300) else RED
+        print("  " + YELLOW(BOLD("<%d" % seq_num)),
+              "[%f]" % time_started,
+              BLUE(method),
+              req_id,
+              url,
+              "<=",
+              code_format(str(response_status)),
+              WHITE(BOLD("(%dms)" % t)),
+              content_to_print,
+              file=sys.stderr)
+
+
+def _test_tls_version():
+    tls12_check_script = os.path.join(os.getenv("DNANEXUS_HOME"), "build", "tls12check.py")
+    if not os.path.exists(tls12_check_script):
+        return
+
     try:
-        seconds_to_wait = int(response.headers.get('retry-after', DEFAULT_RETRY_AFTER_503_INTERVAL))
-    except ValueError:
-        # retry-after could be formatted as absolute time
-        # instead of seconds to wait. We don't know how to
-        # parse that, but the apiserver doesn't generate
-        # such responses anyway.
-        seconds_to_wait = DEFAULT_RETRY_AFTER_503_INTERVAL
-    return max(1, seconds_to_wait)
+        subprocess.check_output(['python', tls12_check_script])
+    except subprocess.CalledProcessError as e:
+        if e.returncode == 1:
+            print (e.output)
+            raise exceptions.InvalidTLSProtocol
 
 
 def DXHTTPRequest(resource, data, method='POST', headers=None, auth=True,
                   timeout=DEFAULT_TIMEOUT,
                   use_compression=None, jsonify_data=True, want_full_response=False,
                   decode_response_body=True, prepend_srv=True, session_handler=None,
-                  max_retries=DEFAULT_RETRIES, always_retry=False, **kwargs):
+                  max_retries=DEFAULT_RETRIES, always_retry=False,
+                  **kwargs):
     '''
     :param resource: API server route, e.g. "/record/new". If *prepend_srv* is False, a fully qualified URL is expected. If this argument is a callable, it will be called just before each request attempt, and expected to return a tuple (URL, headers). Headers returned by the callback are updated with *headers* (including headers set by this method).
     :type resource: string
@@ -344,23 +540,10 @@ def DXHTTPRequest(resource, data, method='POST', headers=None, auth=True,
 
     global _UPGRADE_NOTIFY
 
+    seq_num = _get_sequence_number()
+
     url = APISERVER + resource if prepend_srv else resource
-    method = method.upper() # Convert method name to uppercase, to ease string comparisons later
-    if _DEBUG >= 3:
-        try:
-            formatted_data = json.dumps(data, indent=2)
-        except UnicodeDecodeError:
-            formatted_data = "<binary data>"
-        print(method, url, "=>\n" + formatted_data, file=sys.stderr)
-    elif _DEBUG == 2:
-        try:
-            formatted_data = json.dumps(data)
-        except UnicodeDecodeError:
-            formatted_data = "<binary data>"
-        print(method, url, "=>", formatted_data, file=sys.stderr)
-    elif _DEBUG > 0:
-        from repr import Repr
-        print(method, url, "=>", Repr().repr(data), file=sys.stderr)
+    method = method.upper()  # Convert method name to uppercase, to ease string comparisons later
 
     if auth is True:
         auth = AUTH_HELPER
@@ -368,10 +551,18 @@ def DXHTTPRequest(resource, data, method='POST', headers=None, auth=True,
     if auth:
         auth(_RequestForAuth(method, url, headers))
 
+    pool_args = {arg: kwargs.pop(arg, None) for arg in ("verify", "cert_file", "key_file")}
+    test_retry = kwargs.pop("_test_retry_http_request", False)
+
+    # data is a sequence/buffer or a dict
+    # serialized_data is a sequence/buffer
+
     if jsonify_data:
-        data = json.dumps(data)
+        serialized_data = json.dumps(data)
         if 'Content-Type' not in headers and method == 'POST':
             headers['Content-Type'] = 'application/json'
+    else:
+        serialized_data = data
 
     # If the input is a buffer, its data gets consumed by
     # requests.request (moving the read position). Record the initial
@@ -381,18 +572,54 @@ def DXHTTPRequest(resource, data, method='POST', headers=None, auth=True,
     if hasattr(data, 'seek') and hasattr(data, 'tell'):
         rewind_input_buffer_offset = data.tell()
 
-    try_index = 0
-    while True:
-        success, streaming_response_truncated, time_started = True, False, None
-        response = None
-        try:
-            if _DEBUG > 0:
-                time_started = time.time()
-            _method, _url, _headers = _process_method_url_headers(method, url, headers)
-            response = _get_pool_manager(kwargs).request(_method, _url, headers=_headers, body=data,
-                                                         timeout=timeout, retries=False, **kwargs)
+    # Maintain two separate counters for the number of tries...
 
-            if _UPGRADE_NOTIFY and response.headers.get('x-upgrade-info', '').startswith('A recommended update is available') and not os.environ.has_key('_ARGCOMPLETE'):
+    try_index = 0  # excluding 503 errors. The number of tries as given here
+                   # cannot exceed (max_retries + 1).
+    try_index_including_503 = 0  # including 503 errors. This number is used to
+                                 # do exponential backoff.
+
+    retried_responses = []
+    _url = None
+    while True:
+        success, time_started = True, None
+        response = None
+        req_id = None
+        try:
+            time_started = time.time()
+            _method, _url, _headers = _process_method_url_headers(method, url, headers)
+
+            _debug_print_request(_DEBUG, seq_num, time_started, _method, _url, _headers, jsonify_data, data)
+
+            body = _maybe_truncate_request(_url, serialized_data)
+
+            # throws BadStatusLine if the server returns nothing
+            try:
+                pool_manager = _get_pool_manager(**pool_args)
+
+                _headers['User-Agent'] = USER_AGENT
+                _headers['DNAnexus-API'] = API_VERSION
+
+                # Converted Unicode headers to ASCII and throw an error if not possible
+                def ensure_ascii(i):
+                    if not isinstance(i, bytes):
+                        i = i.encode('ascii')
+                    return i
+
+                _headers = {ensure_ascii(k): ensure_ascii(v) for k, v in _headers.items()}
+                response = pool_manager.request(_method, _url, headers=_headers, body=body,
+                                                timeout=timeout, retries=False, **kwargs)
+            except urllib3.exceptions.ClosedPoolError:
+                # If another thread closed the pool before the request was
+                # started, will throw ClosedPoolError
+                raise exceptions.UrllibInternalError("ClosedPoolError")
+
+            _raise_error_for_testing(try_index, method)
+            req_id = response.headers.get("x-request-id", "unavailable")
+
+            if (_UPGRADE_NOTIFY
+               and response.headers.get('x-upgrade-info', '').startswith('A recommended update is available')
+               and '_ARGCOMPLETE' not in os.environ):
                 logger.info(response.headers['x-upgrade-info'])
                 try:
                     with file(_UPGRADE_NOTIFY, 'a'):
@@ -406,13 +633,30 @@ def DXHTTPRequest(resource, data, method='POST', headers=None, auth=True,
             if response.status // 100 != 2:
                 # response.headers key lookup is case-insensitive
                 if response.headers.get('content-type', '').startswith('application/json'):
-                    content = json.loads(response.data.decode('utf-8'))
+                    try:
+                        content = response.data.decode('utf-8')
+                    except AttributeError:
+                        raise exceptions.UrllibInternalError("Content is none", response.status)
+                    try:
+                        content = json.loads(content)
+                    except ValueError:
+                        # The JSON is not parsable, but we should be able to retry.
+                        raise exceptions.BadJSONInReply("Invalid JSON received from server", response.status)
                     try:
                         error_class = getattr(exceptions, content["error"]["type"], exceptions.DXAPIError)
                     except (KeyError, AttributeError, TypeError):
-                        error_class = exceptions.HTTPError
-                    raise error_class(content, response.status)
-                raise exceptions.HTTPError("{} {}".format(response.status, response.reason))
+                        raise exceptions.HTTPError(response.status, content)
+                    raise error_class(content, response.status, time_started, req_id)
+                else:
+                    try:
+                        content = response.data.decode('utf-8')
+                    except AttributeError:
+                        raise exceptions.UrllibInternalError("Content is none", response.status)
+                    raise exceptions.HTTPError("{} {} [Time={} RequestID={}]\n{}".format(response.status,
+                                                                                         response.reason,
+                                                                                         time_started,
+                                                                                         req_id,
+                                                                                         content))
 
             if want_full_response:
                 return response
@@ -421,91 +665,124 @@ def DXHTTPRequest(resource, data, method='POST', headers=None, auth=True,
                     if int(response.headers['content-length']) != len(response.data):
                         range_str = (' (%s)' % (headers['Range'],)) if 'Range' in headers else ''
                         raise exceptions.ContentLengthError(
-                            "Received response with content-length header set to %s but content length is %d%s" %
-                            (response.headers['content-length'], len(response.data), range_str)
+                            "Received response with content-length header set to %s but content length is %d%s. " +
+                            "[Time=%f RequestID=%s]" %
+                            (response.headers['content-length'], len(response.data), range_str, time_started, req_id)
                         )
 
                 content = response.data
 
+                response_was_json = False
+
                 if decode_response_body:
                     content = content.decode('utf-8')
                     if response.headers.get('content-type', '').startswith('application/json'):
                         try:
                             content = json.loads(content)
-                            if _DEBUG > 0:
-                                t = int((time.time() - time_started) * 1000)
-                            if _DEBUG >= 3:
-                                print(method, url, "<=", response.status, "(%dms)" % t,
-                                      "\n" + json.dumps(content, indent=2), file=sys.stderr)
-                            elif _DEBUG == 2:
-                                print(method, url, "<=", response.status, "(%dms)" % t, json.dumps(content),
-                                      file=sys.stderr)
-                            elif _DEBUG > 0:
-                                print(method, url, "<=", response.status, "(%dms)" % t, Repr().repr(content),
-                                      file=sys.stderr)
                         except ValueError:
-                            # If a streaming API call (no content-length
-                            # set) encounters an error it may just halt the
-                            # response because it has no other way to
-                            # indicate an error. Under these circumstances
-                            # the client sees unparseable JSON, and we
-                            # should be able to recover.
-                            streaming_response_truncated = 'content-length' not in response.headers
-                            raise exceptions.HTTPError("Invalid JSON received from server")
+                            # The JSON is not parsable, but we should be able to retry.
+                            raise exceptions.BadJSONInReply("Invalid JSON received from server", response.status)
+                        else:
+                            response_was_json = True
+
+                req_id = response.headers.get('x-request-id') or "--"
+
+                _debug_print_response(_DEBUG, seq_num, time_started, req_id, response.status, response_was_json,
+                                      _method, _url, content)
+
+                if test_retry:
+                    retried_responses.append(content)
+                    if len(retried_responses) == 1:
+                        continue
+                    else:
+                        _set_retry_response(retried_responses[0])
+                        return retried_responses[1]
+
                 return content
             raise AssertionError('Should never reach this line: expected a result to have been returned by now')
         except Exception as e:
             # Avoid reusing connections in the pool, since they may be
             # in an inconsistent state (observed as "ResponseNotReady"
             # errors).
-            _get_pool_manager(kwargs).clear()
+            _get_pool_manager(**pool_args).clear()
             success = False
             exception_msg = _extract_msg_from_last_exception()
             if isinstance(e, _expected_exceptions):
-                if response is not None and response.status == 503:
-                    seconds_to_wait = _extract_retry_after_timeout(response)
-                    logger.warn("%s %s: %s. Waiting %d seconds due to server unavailability...",
-                                method, url, exception_msg, seconds_to_wait)
-                    time.sleep(seconds_to_wait)
-                    # Note, we escape the "except" block here without
-                    # incrementing try_index because 503 responses with
-                    # Retry-After should not count against the number of
-                    # permitted retries.
-                    continue
-
-                # Total number of allowed tries is the initial try + up to
-                # (max_retries) subsequent retries.
+                # Total number of allowed tries is the initial try PLUS
+                # up to (max_retries) subsequent retries.
                 total_allowed_tries = max_retries + 1
                 ok_to_retry = False
-                # Because try_index is not incremented until we escape this
-                # iteration of the loop, try_index is equal to the number of
-                # tries that have failed so far, minus one. Test whether we
-                # have exhausted all retries.
+                is_retryable = always_retry or (method == 'GET') or _is_retryable_exception(e)
+                # Because try_index is not incremented until we escape
+                # this iteration of the loop, try_index is equal to the
+                # number of tries that have failed so far, minus one.
                 if try_index + 1 < total_allowed_tries:
-                    if response is None or isinstance(e, exceptions.ContentLengthError) or \
-                       streaming_response_truncated:
-                        ok_to_retry = always_retry or (method == 'GET') or _is_retryable_exception(e)
+                    # BadStatusLine ---  server did not return anything
+                    # BadJSONInReply --- server returned JSON that didn't parse properly
+                    if (response is None
+                       or isinstance(e, (exceptions.ContentLengthError, BadStatusLine, exceptions.BadJSONInReply,
+                                         urllib3.exceptions.ProtocolError, exceptions.UrllibInternalError))):
+                        ok_to_retry = is_retryable
                     else:
                         ok_to_retry = 500 <= response.status < 600
 
+                    # The server has closed the connection prematurely
+                    if (response is not None
+                       and response.status == 400 and is_retryable and method == 'PUT'
+                       and isinstance(e, requests.exceptions.HTTPError)):
+                        if '<Code>RequestTimeout</Code>' in exception_msg:
+                            logger.info("Retrying 400 HTTP error, due to slow data transfer. " +
+                                        "Request Time=%f Request ID=%s", time_started, req_id)
+                        else:
+                            logger.info("400 HTTP error, of unknown origin, exception_msg=[%s]. " +
+                                        "Request Time=%f Request ID=%s", exception_msg, time_started, req_id)
+                        ok_to_retry = True
+
+                    # Unprocessable entity, request has semantical errors
+                    if response is not None and response.status == 422:
+                        ok_to_retry = False
+
                 if ok_to_retry:
                     if rewind_input_buffer_offset is not None:
                         data.seek(rewind_input_buffer_offset)
-                    delay = min(2 ** try_index, DEFAULT_TIMEOUT)
-                    logger.warn("%s %s: %s. Waiting %d seconds before retry %d of %d...",
-                                method, url, exception_msg, delay, try_index + 1, max_retries)
+
+                    delay = _calculate_retry_delay(response, try_index_including_503 + 1)
+
+                    range_str = (' (range=%s)' % (headers['Range'],)) if 'Range' in headers else ''
+                    if response is not None and response.status == 503:
+                        waiting_msg = 'Waiting %d seconds before retry...' % (delay,)
+                    else:
+                        waiting_msg = 'Waiting %d seconds before retry %d of %d...' % (
+                            delay, try_index + 1, max_retries)
+
+                    logger.warn("[%s] %s %s: %s. %s %s",
+                                time.ctime(), method, _url, exception_msg, waiting_msg, range_str)
                     time.sleep(delay)
-                    try_index += 1
+                    try_index_including_503 += 1
+                    if response is None or response.status != 503:
+                        try_index += 1
                     continue
 
             # All retries have been exhausted OR the error is deemed not
             # retryable. Print the latest error and propagate it back to the caller.
             if not isinstance(e, exceptions.DXAPIError):
-                logger.error("%s %s: %s", method, url, exception_msg)
+                logger.error("[%s] %s %s: %s.", time.ctime(), method, _url, exception_msg)
+
+            if isinstance(e, urllib3.exceptions.ProtocolError) and \
+                'Connection reset by peer' in exception_msg:
+                # If the protocol error is 'connection reset by peer', most likely it is an
+                # error in the ssl handshake due to unsupported TLS protocol.
+                _test_tls_version()
+
+            # Retries have been exhausted, and we are unable to get a full
+            # buffer from the data source. Raise a special exception.
+            if isinstance(e, urllib3.exceptions.ProtocolError) and \
+               'Connection broken: IncompleteRead' in exception_msg:
+                raise exceptions.DXIncompleteReadsError(exception_msg)
             raise
         finally:
             if success and try_index > 0:
-                logger.info("%s %s: Recovered after %d retries", method, url, try_index)
+                logger.info("[%s] %s %s: Recovered after %d retries", time.ctime(), method, _url, try_index)
 
         raise AssertionError('Should never reach this line: should have attempted a retry or reraised by now')
     raise AssertionError('Should never reach this line: should never break out of loop')
@@ -523,6 +800,49 @@ class DXHTTPOAuth2(AuthBase):
             raise NotImplementedError("Token types other than bearer are not yet supported")
         return r
 
+
+'''
+This function is used for reading a part of an S3 object. It returns a string containing the data. If there is an
+error, and exception is thrown.
+
+There is special handling if a DXIncompleteReadsError is thrown, for which urllib3 gets only part of the requested
+range from the chunk of data. The range is split into smaller chunks, and each sub-chunk is tried in a DXHTTPRequest.
+The smaller chunks are then concatenated to form the original range of data. If a DXIncompleteReadsError is thrown
+(after retrying the sub-chunk 6 times) while reading a sub-chunk, then we fail.
+'''
+
+
+def _dxhttp_read_range(url, headers, start_pos, end_pos, timeout, sub_range=True):
+    if sub_range:
+        headers['Range'] = "bytes=" + str(start_pos) + "-" + str(end_pos)
+    try:
+        data = DXHTTPRequest(url, '', method='GET', headers=headers, auth=None, jsonify_data=False, prepend_srv=False,
+                             always_retry=True, timeout=timeout, decode_response_body=False)
+        _raise_error_for_testing()
+        return data
+
+    # When chunk fails to be read, it gets broken into sub-chunks
+    except exceptions.DXIncompleteReadsError:
+        chunk_buffer = StringIO()
+        subchunk_len = int(math.ceil((end_pos - start_pos + 1)/INCOMPLETE_READS_NUM_SUBCHUNKS))
+        subchunk_start_pos = start_pos
+
+        while subchunk_start_pos <= end_pos:
+            subchunk_end_pos = min(subchunk_start_pos + subchunk_len - 1, end_pos)
+            headers['Range'] = "bytes=" + str(subchunk_start_pos) + "-" + str(subchunk_end_pos)
+            subchunk_start_pos += subchunk_len
+            data = DXHTTPRequest(url, '', method='GET', headers=headers, auth=None, jsonify_data=False,
+                                 prepend_srv=False, always_retry=True, timeout=timeout,
+                                 decode_response_body=False)
+
+            # Concatenate sub-chunks
+            chunk_buffer.write(data)
+
+        concat_chunks = chunk_buffer.getvalue()
+        chunk_buffer.close()
+        return concat_chunks
+
+
 def set_api_server_info(host=None, port=None, protocol=None):
     '''
     :param host: API server hostname
@@ -609,7 +929,7 @@ def set_project_context(dxid):
     global PROJECT_CONTEXT_ID
     PROJECT_CONTEXT_ID = dxid
 
-def get_auth_server_name(host_override=None, port_override=None):
+def get_auth_server_name(host_override=None, port_override=None, protocol='https'):
     """
     Chooses the auth server name from the currently configured API server name.
 
@@ -619,11 +939,15 @@ def get_auth_server_name(host_override=None, port_override=None):
     if host_override is not None or port_override is not None:
         if host_override is None or port_override is None:
             raise exceptions.DXError("Both host and port must be specified if either is specified")
-        return 'http://' + host_override + ':' + str(port_override)
+        return protocol + '://' + host_override + ':' + str(port_override)
     elif APISERVER_HOST == 'stagingapi.dnanexus.com':
         return 'https://stagingauth.dnanexus.com'
     elif APISERVER_HOST == 'api.dnanexus.com':
         return 'https://auth.dnanexus.com'
+    elif APISERVER_HOST == 'stagingapi.cn.dnanexus.com':
+        return 'https://stagingauth.cn.dnanexus.com:7001'
+    elif APISERVER_HOST == 'api.cn.dnanexus.com':
+        return 'https://auth.cn.dnanexus.com:8001'
     elif APISERVER_HOST == "localhost" or APISERVER_HOST == "127.0.0.1":
         if "DX_AUTHSERVER_HOST" not in os.environ or "DX_AUTHSERVER_PORT" not in os.environ:
             err_msg = "Must set authserver env vars (DX_AUTHSERVER_HOST, DX_AUTHSERVER_PORT) if apiserver is {apiserver}."
@@ -634,6 +958,42 @@ def get_auth_server_name(host_override=None, port_override=None):
         err_msg = "Could not determine which auth server is associated with {apiserver}."
         raise exceptions.DXError(err_msg.format(apiserver=APISERVER_HOST))
 
+
+'''This field is used for testing a retry of an Http request. The caller can pass
+an argument "_test_retry_http_request"=1 to DXHTTPREQUEST to simulate a request that
+required a retry. The first response will be returned and the second response can be
+retrieved by calling _get_retry_response
+'''
+
+_retry_response = None
+
+
+def _set_retry_response(response):
+    global _retry_response
+    _retry_response = response
+
+
+def _get_retry_response():
+    return _retry_response
+
+def append_underlying_workflow_describe(globalworkflow_desc):
+    """
+    Adds the "workflowDescribe" field to the config for each region of
+    the global workflow. The value is the description of an underlying
+    workflow in that region.
+    """
+    if not globalworkflow_desc or \
+            globalworkflow_desc['class'] != 'globalworkflow' or \
+            not 'regionalOptions' in globalworkflow_desc:
+        return globalworkflow_desc
+
+    for region, config in globalworkflow_desc['regionalOptions'].items():
+        workflow_id = config['workflow']
+        workflow_desc = dxpy.api.workflow_describe(workflow_id)
+        globalworkflow_desc['regionalOptions'][region]['workflowDescribe'] = workflow_desc
+    return globalworkflow_desc
+
+
 from .utils.config import DXConfig as _DXConfig
 config = _DXConfig()
 
