@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2015 DNAnexus, Inc.
+# Copyright (C) 2013-2016 DNAnexus, Inc.
 #
 # This file is part of dx-toolkit (DNAnexus platform client libraries).
 #
@@ -56,6 +56,18 @@ from ..compat import basestring
 # DXApp #
 #########
 
+_app_required_keys = ['name', 'title', 'summary', 'dxapi', 'openSource',
+                      'version', 'inputSpec', 'outputSpec', 'runSpec',
+                      'developers', 'authorizedUsers', 'regionalOptions']
+
+# These are optional keys for apps, not sure what to do with them
+_app_optional_keys = ['details', 'categories', 'access']
+
+_app_describe_output_keys = []
+
+_app_cleanup_keys = ['name', 'title', 'summary', 'dxapi', 'openSource',
+                     'version', 'runSpec', 'developers', 'authorizedUsers']
+
 class DXApp(DXObject, DXExecutable):
     '''
     Remote app object handler.
@@ -160,6 +172,7 @@ class DXApp(DXObject, DXExecutable):
         is called.
 
         '''
+        #TODO: add support for regionalOptions (and deprecate top-level applet and resources)
         dx_hash = {}
         if 'applet' not in kwargs and 'initializeFrom' not in kwargs:
             raise DXError("%s: One of the keyword arguments %s and %s is required" % (self.__class__.__name__, 'applet', 'initializeFrom'))
@@ -353,6 +366,18 @@ class DXApp(DXObject, DXExecutable):
         # May need to be changed when workflow apps are enabled
         return DXExecutable._get_run_input_fields_for_applet(executable_input, **kwargs)
 
+    def _get_required_keys(self):
+        return _app_required_keys
+
+    def _get_optional_keys(self):
+        return _app_optional_keys
+
+    def _get_describe_output_keys(self):
+        return _app_describe_output_keys
+
+    def _get_cleanup_keys(self):
+        return _app_cleanup_keys
+
     def run(self, app_input, *args, **kwargs):
         """
         Creates a new job that executes the function "main" of this app with
