@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2015 DNAnexus, Inc.
+# Copyright (C) 2013-2016 DNAnexus, Inc.
 #
 # This file is part of dx-toolkit (DNAnexus platform client libraries).
 #
@@ -24,18 +24,21 @@ This remote file handler is a Python file-like object.
 from __future__ import print_function, unicode_literals, division, absolute_import
 
 import os, sys, logging, traceback, hashlib, copy, time
-import concurrent.futures
+import math
+import mmap
+from threading import Lock
 from multiprocessing import cpu_count
 
 import dxpy
 from . import DXDataObject
-from ..exceptions import DXFileError
+from ..exceptions import DXFileError, DXIncompleteReadsError
 from ..utils import warn
 from ..utils.resolver import object_exists_in_project
-from ..compat import BytesIO
+from ..compat import BytesIO, basestring
 
 
 DXFILE_HTTP_THREADS = min(cpu_count(), 8)
+MIN_BUFFER_SIZE = 1024*1024
 DEFAULT_BUFFER_SIZE = 1024*1024*16
 if dxpy.JOB_ID:
     # Increase HTTP request buffer size when we are running within the
@@ -45,6 +48,86 @@ if dxpy.JOB_ID:
 MD5_READ_CHUNK_SIZE = 1024*1024*4
 FILE_REQUEST_TIMEOUT = 60
 
+
+def _validate_headers(headers):
+    for key, value in headers.items():
+        if not isinstance(key, basestring):
+            raise ValueError("Expected key %r of headers to be a string" % (key,))
+        if not isinstance(value, basestring):
+            raise ValueError("Expected value %r of headers (associated with key %r) to be a string"
+                             % (value, key))
+    return headers
+
+
+def _readable_part_size(num_bytes):
+    "Returns the file size in readable form."
+    B = num_bytes
+    KB = float(1024)
+    MB = float(KB * 1024)
+    GB = float(MB * 1024)
+    TB = float(GB * 1024)
+
+    if B < KB:
+        return '{0} {1}'.format(B, 'bytes' if B != 1 else 'byte')
+    elif KB <= B < MB:
+        return '{0:.2f} KiB'.format(B/KB)
+    elif MB <= B < GB:
+        return '{0:.2f} MiB'.format(B/MB)
+    elif GB <= B < TB:
+        return '{0:.2f} GiB'.format(B/GB)
+    elif TB <= B:
+        return '{0:.2f} TiB'.format(B/TB)
+
+
+def _get_write_buf_size(buffer_size_hint, file_upload_params, expected_file_size, file_is_mmapd=False):
+    max_num_parts = file_upload_params['maximumNumParts']
+    min_part_size = file_upload_params['minimumPartSize']
+    max_part_size = file_upload_params['maximumPartSize']
+    max_file_size = file_upload_params['maximumFileSize']
+
+    if expected_file_size is not None and expected_file_size > max_file_size:
+        raise DXFileError("Size of file exceeds maximum of {}".format(_readable_part_size(max_file_size)))
+
+    min_buffer_size = min_part_size
+    if expected_file_size is not None:
+        # Raise buffer size (for files exceeding DEFAULT_BUFFER_SIZE
+        # * the maximium parts allowed bytes) in order to prevent us
+        # from exceeding the configured parts limit.
+        min_buffer_size = max(min_buffer_size, int(math.ceil(float(expected_file_size) / max_num_parts)))
+    max_buffer_size = max_part_size
+
+    assert min_buffer_size <= max_buffer_size
+
+    if file_is_mmapd:
+        # If file is mmapd, force the eventual result to be a
+        # multiple of the allocation granularity by rounding all of
+        # buffer_size, min_buffer_size, and max_buffer_size to a
+        # nearby multiple of the allocation granularity (below, the
+        # final buffer size will be one of these).
+        if min_buffer_size % mmap.ALLOCATIONGRANULARITY != 0:
+            min_buffer_size += mmap.ALLOCATIONGRANULARITY - min_buffer_size % mmap.ALLOCATIONGRANULARITY
+        if max_buffer_size % mmap.ALLOCATIONGRANULARITY != 0:
+            max_buffer_size -= max_buffer_size % mmap.ALLOCATIONGRANULARITY
+        buffer_size_hint = buffer_size_hint - buffer_size_hint % mmap.ALLOCATIONGRANULARITY
+    else:
+        buffer_size_hint = buffer_size_hint
+
+    # Use the user-specified hint if it is a permissible size
+    # (satisfies API and large enough to upload file of advertised
+    # size). Otherwise, select the closest size that is permissible.
+    buffer_size = buffer_size_hint
+    buffer_size = max(buffer_size, min_buffer_size)
+    buffer_size = min(buffer_size, max_buffer_size)
+
+    if expected_file_size is not None and (buffer_size * max_num_parts < expected_file_size):
+        raise AssertionError("part size would be too small to upload the requested number of bytes")
+
+    if file_is_mmapd and buffer_size % mmap.ALLOCATIONGRANULARITY != 0:
+        raise AssertionError('part size will not be accepted by mmap')
+
+    return buffer_size
+
+
 class DXFile(DXDataObject):
     '''Remote file object handler.
 
@@ -85,22 +168,44 @@ class DXFile(DXDataObject):
     _close = staticmethod(dxpy.api.file_close)
     _list_projects = staticmethod(dxpy.api.file_list_projects)
 
-    _http_threadpool = None
     _http_threadpool_size = DXFILE_HTTP_THREADS
+    _http_threadpool = dxpy.utils.get_futures_threadpool(max_workers=_http_threadpool_size)
 
     NO_PROJECT_HINT = 'NO_PROJECT_HINT'
 
     @classmethod
     def set_http_threadpool_size(cls, num_threads):
-        cls._http_threadpool_size = num_threads
+        '''
 
-    @classmethod
-    def _ensure_http_threadpool(cls):
-        if cls._http_threadpool is None:
-            cls._http_threadpool = dxpy.utils.get_futures_threadpool(max_workers=cls._http_threadpool_size)
+        .. deprecated:: 0.191.0
 
-    def __init__(self, dxid=None, project=None, mode=None,
-                 read_buffer_size=DEFAULT_BUFFER_SIZE, write_buffer_size=DEFAULT_BUFFER_SIZE):
+        '''
+        print('set_http_threadpool_size is deprecated')
+
+    def __init__(self, dxid=None, project=None, mode=None, read_buffer_size=DEFAULT_BUFFER_SIZE,
+                 write_buffer_size=DEFAULT_BUFFER_SIZE, expected_file_size=None, file_is_mmapd=False):
+        """
+        :param dxid: Object ID
+        :type dxid: string
+        :param project: Project ID
+        :type project: string
+        :param mode: One of "r", "w", or "a" for read, write, and append
+            modes, respectively
+        :type mode: string
+        :param read_buffer_size: size of read buffer in bytes
+        :type read_buffer_size: int
+        :param write_buffer_size: hint for size of write buffer in
+            bytes. A lower or higher value may be used depending on
+            region-specific parameters and on the expected file size.
+        :type write_buffer_size: int
+        :param expected_file_size: size of data that will be written, if
+            known
+        :type expected_file_size: int
+        :param file_is_mmapd: True if input file is mmap'd (if so, the
+            write buffer size will be constrained to be a multiple of
+            the allocation granularity)
+        :type file_is_mmapd: bool
+        """
         DXDataObject.__init__(self, dxid=dxid, project=project)
         if mode is None:
             self._close_on_exit = True
@@ -108,17 +213,28 @@ class DXFile(DXDataObject):
             if mode not in ['r', 'w', 'a']:
                 raise ValueError("mode must be one of 'r', 'w', or 'a'")
             self._close_on_exit = (mode == 'w')
-
         self._read_buf = BytesIO()
         self._write_buf = BytesIO()
 
-        if write_buffer_size < 5*1024*1024:
-            raise DXFileError("Write buffer size must be at least 5 MB")
-
         self._read_bufsize = read_buffer_size
-        self._write_bufsize = write_buffer_size
 
+        # Computed lazily later since this depends on the project, and
+        # we want to allow the project to be set as late as possible.
+        # Call _ensure_write_bufsize to ensure that this is set before
+        # trying to read it.
+        self._write_bufsize = None
+
+        self._write_buffer_size_hint = write_buffer_size
+        self._expected_file_size = expected_file_size
+        self._file_is_mmapd = file_is_mmapd
+
+        # These are cached once for all download threads. This saves calls to the apiserver.
         self._download_url, self._download_url_headers, self._download_url_expires = None, None, None
+
+        # This lock protects accesses to the above three variables, ensuring that they would
+        # be checked and changed atomically. This protects against thread race conditions.
+        self._url_download_mutex = Lock()
+
         self._request_iterator, self._response_iterator = None, None
         self._http_threadpool_futures = set()
 
@@ -205,6 +321,9 @@ class DXFile(DXDataObject):
         if _buffer:
             yield _buffer
 
+    next = next
+    __next__ = next
+
     def set_ids(self, dxid, project=None):
         '''
         :param dxid: Object ID
@@ -320,8 +439,6 @@ class DXFile(DXDataObject):
                 self._http_threadpool_futures = set()
 
     def _async_upload_part_request(self, *args, **kwargs):
-        self._ensure_http_threadpool()
-
         while len(self._http_threadpool_futures) >= self._http_threadpool_size:
             future = dxpy.utils.wait_for_a_future(self._http_threadpool_futures)
             if future.exception() != None:
@@ -331,10 +448,26 @@ class DXFile(DXDataObject):
         future = self._http_threadpool.submit(self.upload_part, *args, **kwargs)
         self._http_threadpool_futures.add(future)
 
+    def _ensure_write_bufsize(self, **kwargs):
+        if self._write_bufsize is not None:
+            return
+        file_upload_params = dxpy.api.project_describe(
+            self.get_proj_id(),
+            {'fields': {'fileUploadParameters': True}},
+            **kwargs
+        )['fileUploadParameters']
+        self._empty_last_part_allowed = file_upload_params['emptyLastPartAllowed']
+        self._write_bufsize = _get_write_buf_size(self._write_buffer_size_hint,
+                                                  file_upload_params,
+                                                  self._expected_file_size,
+                                                  self._file_is_mmapd)
+
     def write(self, data, multithread=True, **kwargs):
         '''
         :param data: Data to be written
         :type data: str or mmap object
+        :param multithread: If True, sends multiple write requests asynchronously
+        :type multithread: boolean
 
         Writes the data *data* to the file.
 
@@ -344,6 +477,7 @@ class DXFile(DXDataObject):
             does not affect where the next :meth:`write` will occur.
 
         '''
+        self._ensure_write_bufsize(**kwargs)
 
         def write_request(data_for_write_req):
             if multithread:
@@ -408,9 +542,14 @@ class DXFile(DXDataObject):
         '''
         self.flush(**kwargs)
 
-        if self._num_uploaded_parts == 0:
-            # We haven't uploaded any parts in this session. In case no parts have been uploaded at all, try to upload
-            # an empty part (files with 0 parts cannot be closed).
+        # Also populates emptyLastPartAllowed
+        self._ensure_write_bufsize(**kwargs)
+
+        if self._num_uploaded_parts == 0 and self._empty_last_part_allowed:
+            # We haven't uploaded any parts in this session.
+            # In case no parts have been uploaded at all and region
+            # settings allow last empty part upload, try to upload
+            # an empty part (otherwise files with 0 parts cannot be closed).
             try:
                 self.upload_part('', 1, **kwargs)
             except dxpy.exceptions.InvalidState:
@@ -451,7 +590,6 @@ class DXFile(DXDataObject):
         defaults to 1. This probably only makes sense if this is the
         only part to be uploaded.
         """
-
         req_input = {}
         if index is not None:
             req_input["index"] = int(index)
@@ -487,7 +625,7 @@ class DXFile(DXDataObject):
 
             resp = dxpy.api.file_upload(self._dxid, req_input, **kwargs)
             url = resp["url"]
-            return url, resp.get("headers", {})
+            return url, _validate_headers(resp.get("headers", {}))
 
         # The file upload API requires us to get a pre-authenticated upload URL (and headers for it) every time we
         # attempt an upload. Because DXHTTPRequest will retry requests under retryable conditions, we give it a callback
@@ -509,10 +647,10 @@ class DXFile(DXDataObject):
         if report_progress_fn is not None:
             report_progress_fn(self, len(data))
 
-    def get_download_url(self, duration=24*3600, preauthenticated=False, filename=None, project=None, **kwargs):
+    def get_download_url(self, duration=None, preauthenticated=False, filename=None, project=None, **kwargs):
         """
         :param duration: number of seconds for which the generated URL will be
-            valid
+            valid, should only be specified when preauthenticated is True
         :type duration: int
         :param preauthenticated: if True, generates a 'preauthenticated'
             download URL, which embeds authentication info in the URL and does
@@ -522,46 +660,92 @@ class DXFile(DXDataObject):
         :type filename: str
         :param project: ID of a project containing the file (the download URL
             will be associated with this project, and this may affect which
-            billing account is billed for this download). If None, no hint is
-            supplied to the API server. If set to a project, the URL is only
-            valid so long as the user has access to that project and the
-            project contains that file.
+            billing account is billed for this download).
+            If no project is specified, an attempt will be made to verify if the file is
+            in the project from the DXFile handler (as specified by the user or
+            the current project stored in dxpy.WORKSPACE_ID). Otherwise, no hint is supplied.
+            This fall back behavior does not happen inside a job environment.
+            A non preauthenticated URL is only valid as long as the user has
+            access to that project and the project contains that file.
         :type project: str
         :returns: download URL and dict containing HTTP headers to be supplied
             with the request
         :rtype: tuple (str, dict)
-        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if the project does
-            not contain this file.
+        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if a project context was
+            given and the file was not found in that project context.
+        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if no project context was
+            given and the file was not found in any projects.
 
         Obtains a URL that can be used to directly download the associated
         file.
 
         """
+        args = {"preauthenticated": preauthenticated}
+
+        if duration is not None:
+            args["duration"] = duration
+        if filename is not None:
+            args["filename"] = filename
+
+        # If project=None, we fall back to the project attached to this handler
+        # (if any). If this is supplied, it's treated as a hint: if it's a
+        # project in which this file exists, it's passed on to the
+        # apiserver. Otherwise, NO hint is supplied. In principle supplying a
+        # project in the handler that doesn't contain this file ought to be an
+        # error, but it's this way for backwards compatibility. We don't know
+        # who might be doing downloads and creating handlers without being
+        # careful that the project encoded in the handler contains the file
+        # being downloaded. They may now rely on such behavior.
+        if project is None and 'DX_JOB_ID' not in os.environ:
+            project_from_handler = self.get_proj_id()
+            if project_from_handler and object_exists_in_project(self.get_id(), project_from_handler):
+                project = project_from_handler
+
+        if project is not None and project is not DXFile.NO_PROJECT_HINT:
+            args["project"] = project
+
         # Test hook to write 'project' argument passed to API call to a
         # local file
         if '_DX_DUMP_BILLED_PROJECT' in os.environ:
             with open(os.environ['_DX_DUMP_BILLED_PROJECT'], "w") as fd:
-                if project is not None:
+                if project is not None and project != DXFile.NO_PROJECT_HINT:
                     fd.write(project)
 
-        args = {"duration": duration, "preauthenticated": preauthenticated}
-        if filename is not None:
-            args["filename"] = filename
-        if project is not None:
-            args["project"] = project
-        if self._download_url is None or self._download_url_expires < time.time():
-            # logging.debug("Download URL unset or expired, requesting a new one")
-            if "timeout" not in kwargs:
-                kwargs["timeout"] = FILE_REQUEST_TIMEOUT
-            resp = dxpy.api.file_download(self._dxid, args, **kwargs)
-            self._download_url = resp["url"]
-            self._download_url_headers = resp.get("headers", {})
-            self._download_url_expires = time.time() + duration - 60 # Try to account for drift
-        return self._download_url, self._download_url_headers
+        with self._url_download_mutex:
+            if self._download_url is None or self._download_url_expires < time.time():
+                # The idea here is to cache a download URL for the entire file, that will
+                # be good for a few minutes. This avoids each thread having to ask the
+                # server for a URL, increasing server load.
+                #
+                # To avoid thread race conditions, this check/update procedure is protected
+                # with a lock.
 
-    def _generate_read_requests(self, start_pos=0, end_pos=None, project=None, **kwargs):
+                # logging.debug("Download URL unset or expired, requesting a new one")
+                if "timeout" not in kwargs:
+                    kwargs["timeout"] = FILE_REQUEST_TIMEOUT
+                resp = dxpy.api.file_download(self._dxid, args, **kwargs)
+                self._download_url = resp["url"]
+                self._download_url_headers = _validate_headers(resp.get("headers", {}))
+                if preauthenticated:
+                    self._download_url_expires = resp["expires"]/1000 - 60  # Try to account for drift
+                else:
+                    self._download_url_expires = 32503680000  # doesn't expire (year 3000)
+
+            # Make a copy, ensuring each thread has its own mutable
+            # version of the headers.  Note: python strings are
+            # immutable, so we can safely give a reference to the
+            # download url.
+            retval_download_url = self._download_url
+            retval_download_url_headers = copy.copy(self._download_url_headers)
+
+        return retval_download_url, retval_download_url_headers
+
+    def _generate_read_requests(self, start_pos=0, end_pos=None, project=None,
+                                limit_chunk_size=None, **kwargs):
         # project=None means no hint is to be supplied to the apiserver. It is
         # an error to supply a project that does not contain this file.
+        if limit_chunk_size is None:
+            limit_chunk_size = self._read_bufsize
 
         if self._file_length == None:
             desc = self.describe(**kwargs)
@@ -572,7 +756,7 @@ class DXFile(DXDataObject):
         if end_pos > self._file_length:
             raise DXFileError("Invalid end_pos")
 
-        def chunk_ranges(start_pos, end_pos, init_chunk_size=1024*64, limit_chunk_size=self._read_bufsize, ramp=2, num_requests_between_ramp=4):
+        def chunk_ranges(start_pos, end_pos, init_chunk_size=1024*64, ramp=2, num_requests_between_ramp=4):
             cur_chunk_start = start_pos
             cur_chunk_size = min(init_chunk_size, limit_chunk_size)
             i = 0
@@ -586,24 +770,16 @@ class DXFile(DXDataObject):
 
         for chunk_start_pos, chunk_end_pos in chunk_ranges(start_pos, end_pos):
             url, headers = self.get_download_url(project=project, **kwargs)
-            headers = copy.copy(headers)
-            headers['Range'] = "bytes=" + str(chunk_start_pos) + "-" + str(chunk_end_pos)
-            yield dxpy.DXHTTPRequest, [url, ''], {'method': 'GET',
-                                                  'headers': headers,
-                                                  'auth': None,
-                                                  'jsonify_data': False,
-                                                  'prepend_srv': False,
-                                                  'always_retry': True,
-                                                  'timeout': FILE_REQUEST_TIMEOUT,
-                                                  'decode_response_body': False}
-
-    def _next_response_content(self):
-        self._ensure_http_threadpool()
+            # It is possible for chunk_end_pos to be outside of the range of the file
+            yield dxpy._dxhttp_read_range, [url, headers, chunk_start_pos, min(chunk_end_pos, self._file_length - 1),
+                                            FILE_REQUEST_TIMEOUT], {}
 
+    def _next_response_content(self, get_first_chunk_sequentially=False):
         if self._response_iterator is None:
             self._response_iterator = dxpy.utils.response_iterator(
                 self._request_iterator,
-                self._http_threadpool
+                self._http_threadpool,
+                do_first_task_sequentially=get_first_chunk_sequentially
             )
         try:
             return next(self._response_iterator)
@@ -628,8 +804,8 @@ class DXFile(DXDataObject):
             project ID.
         :type project: str or None
         :rtype: string
-        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if *project* is
-            supplied and it does not contain this file
+        :raises: :exc:`~dxpy.exceptions.ResourceNotFound` if *project* is supplied
+           and it does not contain this file
 
         Returns the next *length* bytes, or all the bytes until the end of file
         (if no *length* is given or there are fewer than *length* bytes left in
@@ -654,10 +830,7 @@ class DXFile(DXDataObject):
         # size for this heuristic so we don't incur the overhead for
         # tiny files (which wouldn't contribute as much to the load
         # anyway).
-        if self._file_length > 128 * 1024 and self._pos == 0 and dxpy.JOB_ID:
-            get_first_chunk_sequentially = True
-        else:
-            get_first_chunk_sequentially = False
+        get_first_chunk_sequentially = (self._file_length > 128 * 1024 and self._pos == 0 and dxpy.JOB_ID)
 
         if self._pos == self._file_length:
             return b""
@@ -665,26 +838,6 @@ class DXFile(DXDataObject):
         if length == None or length > self._file_length - self._pos:
             length = self._file_length - self._pos
 
-        # Project specified explicitly to this method read(project=...) is
-        # treated strictly. If supplied, it must be a project in which this
-        # file exists. Otherwise, it's an error.
-        #
-        # If project=None, we fall back to the project attached to this handler
-        # (if any). If this is supplied, it's treated as a hint: if it's a
-        # project in which this file exists, it's passed on to the
-        # apiserver. Otherwise, NO hint is supplied. In principle supplying a
-        # project in the handler that doesn't contain this file ought to be an
-        # error, but it's this way for backwards compatibility. We don't know
-        # who might be doing downloads and creating handlers without being
-        # careful that the project encoded in the handler contains the file
-        # being downloaded. They may now rely on such behavior.
-        if project is None:
-            project_from_handler = self.get_proj_id()
-            if project_from_handler and object_exists_in_project(self.get_id(), project_from_handler):
-                project = project_from_handler
-        elif project == DXFile.NO_PROJECT_HINT:
-            project = None
-
         buf = self._read_buf
         buf_remaining_bytes = dxpy.utils.string_buffer_length(buf) - buf.tell()
         if length <= buf_remaining_bytes:
@@ -702,18 +855,7 @@ class DXFile(DXDataObject):
                     self._request_iterator = self._generate_read_requests(
                         start_pos=self._pos, project=project, **kwargs)
 
-                if get_first_chunk_sequentially:
-                    # Make the first chunk request without using the
-                    # usual thread pool and block until it completes. On
-                    # the second chunk, we'll call
-                    # _next_response_content in the alternative block
-                    # below. This starts the threadpool going for the
-                    # second and all subsequent chunks.
-                    _callable, _args, _kwargs = next(self._request_iterator)
-                    content = _callable(*_args, **_kwargs)
-                    get_first_chunk_sequentially = False
-                else:
-                    content = self._next_response_content()
+                content = self._next_response_content(get_first_chunk_sequentially=get_first_chunk_sequentially)
 
                 if len(content) < remaining_len:
                     buf.write(content)
