@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2015 DNAnexus, Inc.
+# Copyright (C) 2013-2016 DNAnexus, Inc.
 #
 # This file is part of dx-toolkit (DNAnexus platform client libraries).
 #
@@ -30,7 +30,6 @@ EXPECTED_ERR_EXIT_STATUS = 3
 
 class DXError(Exception):
     '''Base class for exceptions in this package.'''
-    pass
 
 class DXAPIError(DXError):
     '''
@@ -38,7 +37,7 @@ class DXAPIError(DXError):
     https://wiki.dnanexus.com/API-Specification-v1.0.0/Protocols#Errors for complete documentation of API errors,
     including those reflected by subclasses of this class.
     '''
-    def __init__(self, content, code):
+    def __init__(self, content, code, timestamp="", req_id=""):
         self.name = content["error"]["type"]
         self.msg = content["error"]["message"]
         if "details" in content["error"]:
@@ -46,10 +45,13 @@ class DXAPIError(DXError):
         else:
             self.details = None
         self.code = code
+        self.timestamp = timestamp
+        self.req_id = req_id
 
     def error_message(self):
         "Returns a one-line description of the error."
         output = self.msg + ", code " + str(self.code)
+        output += ". Request Time={}, Request ID={}".format(self.timestamp, self.req_id)
         if self.name != self.__class__.__name__:
             output = self.name + ": " + output
         return output
@@ -62,52 +64,43 @@ class DXAPIError(DXError):
 
 class MalformedJSON(DXAPIError):
     ''' Raised when the input could not be parsed as JSON. '''
-    pass
 
 class InvalidAuthentication(DXAPIError):
     ''' Raised when the provided OAuth2 token is invalid. '''
-    pass
 
 class PermissionDenied(DXAPIError):
     ''' Raised when the supplied credentials have insufficient permissions to perform this action. '''
-    pass
 
 class SpendingLimitExceeded(DXAPIError):
     ''' Raised when the spending limit has been reached for the account that would be billed for this action. '''
-    pass
 
 class ResourceNotFound(DXAPIError):
     ''' Raised when a specified entity or resource could not be found. '''
-    pass
 
 class InvalidInput(DXAPIError):
     ''' Raised when the input is syntactically correct (JSON), but semantically incorrect (for example, a JSON array
     is provided where a hash was required; or a required parameter was missing, etc.). '''
-    pass
 
 class InvalidState(DXAPIError):
     ''' Raised when the operation is not allowed at this object state. '''
-    pass
 
 class InvalidType(DXAPIError):
     ''' Raised when an object specified in the request is of invalid type. '''
-    pass
 
 class RateLimitConditional(DXAPIError):
     ''' Raised when the rate of invalid requests is too high. '''
-    pass
 
 class InternalError(DXAPIError):
     ''' Raised when the server encountered an internal error. '''
-    pass
 
 class ServiceUnavailable(DXAPIError):
     ''' Raised when an API service was temporarily unavailable. '''
-    pass
 
 class DXFileError(DXError):
     '''Exception for :class:`dxpy.bindings.dxfile.DXFile`.'''
-    pass
+
+class DXIncompleteReadsError(DXError):
+    '''Exception for :class:`dxpy.bindings.dxfile.DXFile` when returned read data is shorter than requested'''
 
 class DXPartLengthMismatchError(DXFileError):
     '''Exception raised by :class:`dxpy.bindings.dxfile.DXFile` on part length mismatch.'''
@@ -117,23 +110,18 @@ class DXChecksumMismatchError(DXFileError):
 
 class DXGTableError(DXError):
     '''Exception for :class:`dxpy.bindings.dxgtable.DXGTable`.'''
-    pass
 
 class DXSearchError(DXError):
     '''Exception for :mod:`dxpy.bindings.search` methods.'''
-    pass
 
 class DXAppletError(DXError):
     '''Exception for :class:`dxpy.bindings.dxapplet.DXApplet`.'''
-    pass
 
 class DXJobFailureError(DXError):
     '''Exception produced by :class:`dxpy.bindings.dxjob.DXJob` when a job fails.'''
-    pass
 
 class ProgramError(DXError):
     '''Deprecated. Use :class:`AppError` instead.'''
-    pass
 
 class AppError(ProgramError):
     '''
@@ -146,7 +134,6 @@ class AppError(ProgramError):
     *job_error.json* in the current working directory, allowing reporting of
     the error state through the DNAnexus API.
     '''
-    pass
 
 class AppInternalError(DXError):
     '''
@@ -159,21 +146,43 @@ class AppInternalError(DXError):
     the current working directory, allowing reporting of the error state
     through the DNAnexus API.
     '''
-    pass
 
 class DXCLIError(DXError):
     '''
     Exception class for generic errors in the command-line client
     '''
-    pass
-
 
 class ContentLengthError(HTTPError):
     '''
-    Will be raised when actual content length received from server does not
+    Raised when actual content length received from the server does not
     match the "Content-Length" header
     '''
-    pass
+
+class BadJSONInReply(ValueError):
+    '''
+    Raised when the server returned invalid JSON in the response body. Possible reasons
+    for this are the network connection breaking, or overload on the server.
+    '''
+
+class InvalidTLSProtocol(DXAPIError):
+    '''
+    Raised when the connection to the server was reset due to an ssl protocol not supported.
+    Only connections with TLS v1.2 will be accepted.
+    '''
+    def __init__(self):
+        pass
+
+    def error_message(self):
+        output = "Please refer to our blog post at https://blog.dnanexus.com/2017-09-23-upgrading-tls/ regarding upgrading to TLS 1.2."
+        return output
+
+    def __str__(self):
+        return self.error_message()
+
+class UrllibInternalError(AttributeError):
+    '''
+    Exception class for AttributeError from urllib3
+    '''
 
 
 def format_exception(e):
