@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2015 DNAnexus, Inc.
+# Copyright (C) 2013-2016 DNAnexus, Inc.
 #
 # This file is part of dx-toolkit (DNAnexus platform client libraries).
 #
@@ -23,12 +23,13 @@ from __future__ import print_function, unicode_literals, division, absolute_impo
 
 import sys
 
-from argcomplete import warn
+from ..packages.argcomplete import warn
 from collections import namedtuple, OrderedDict
 import dxpy
-from dxpy.utils.resolver import (get_first_pos_of_char, get_last_pos_of_char, clean_folder_path, resolve_path,
-                                 split_unescaped, ResolutionError)
-from dxpy.utils.printing import fill
+from .resolver import (get_first_pos_of_char, get_last_pos_of_char, clean_folder_path, resolve_path,
+                       split_unescaped, ResolutionError)
+from .printing import fill
+from ..compat import str
 
 def startswith(text):
     return (lambda string: string.startswith(text))
@@ -152,7 +153,7 @@ def get_data_matches(text, delim_pos, dxproj, folderpath, classname=None,
                                               visibility=visibility,
                                               classname=classname,
                                               limit=100,
-                                              describe=True,
+                                              describe=dict(fields=dict(name=True)),
                                               typename=typespec))
         prefix = '' if text == '' else text[:delim_pos + 1]
         return [prefix + escape_name(result['describe']['name']) for result in results]
@@ -231,7 +232,7 @@ def path_completer(text, expected=None, classes=None, perm_level=None,
         try:
             proj_ids, folderpath, entity_name = resolve_path(text, multi_projects=True)
         except ResolutionError as details:
-            sys.stderr.write("\n" + fill(unicode(details)))
+            sys.stderr.write("\n" + fill(str(details)))
             return matches
         for proj in proj_ids:
             # protects against dxpy.WORKSPACE_ID being garbage
@@ -334,7 +335,7 @@ class LocalCompleter():
         self.matches = []
 
     def _populate_matches(self, prefix):
-        from argcomplete.completers import FilesCompleter
+        from ..packages.argcomplete.completers import FilesCompleter
         completer = FilesCompleter()
         self.matches = completer(prefix)
 
@@ -417,24 +418,19 @@ class MultiCompleter():
 
 class InstanceTypesCompleter():
     InstanceTypeSpec = namedtuple('InstanceTypeSpec', ('Name', 'Memory_GB', 'Storage_GB', 'CPU_Cores'))
-    instance_types = OrderedDict()
+
+    # AWS
+    aws_preferred_instance_types = OrderedDict()
     for i in (InstanceTypeSpec('mem1_ssd1_x2', 3.8, 32, 2),
               InstanceTypeSpec('mem1_ssd1_x4', 7.5, 80, 4),
               InstanceTypeSpec('mem1_ssd1_x8', 15.0, 160, 8),
               InstanceTypeSpec('mem1_ssd1_x16', 30.0, 320, 16),
               InstanceTypeSpec('mem1_ssd1_x32', 60.0, 640, 32),
 
-              InstanceTypeSpec('mem1_hdd2_x8', 7.0, 1680, 8),
-              InstanceTypeSpec('mem1_hdd2_x32', 60.5, 3360, 32),
-
               InstanceTypeSpec('mem2_ssd1_x2', 7.5, 32, 2),
               InstanceTypeSpec('mem2_ssd1_x4', 15.0, 80, 4),
               InstanceTypeSpec('mem2_ssd1_x8', 30.0, 160, 8),
 
-              InstanceTypeSpec('mem2_hdd2_x1', 3.8, 410, 1),
-              InstanceTypeSpec('mem2_hdd2_x2', 7.5, 840, 2),
-              InstanceTypeSpec('mem2_hdd2_x4', 15.0, 1680, 4),
-
               InstanceTypeSpec('mem3_ssd1_x2', 15.0, 32, 2),
               InstanceTypeSpec('mem3_ssd1_x4', 30.5, 80, 4),
               InstanceTypeSpec('mem3_ssd1_x8', 61.0, 160, 8),
@@ -445,12 +441,50 @@ class InstanceTypesCompleter():
               InstanceTypeSpec('mem1_ssd2_x4', 7.5, 320, 4),
               InstanceTypeSpec('mem1_ssd2_x8', 15, 640, 8),
               InstanceTypeSpec('mem1_ssd2_x16', 30, 1280, 16),
-              InstanceTypeSpec('mem1_ssd2_x36', 60, 2880, 36),
+              InstanceTypeSpec('mem1_ssd2_x36', 60, 2880, 36)):
+        aws_preferred_instance_types[i.Name] = i
+
+    # Azure
+    azure_preferred_instance_types = OrderedDict()
+    for i in (InstanceTypeSpec('azure:mem1_ssd1_x2', 3.9, 32, 2),
+              InstanceTypeSpec('azure:mem1_ssd1_x4', 7.8, 64, 4),
+              InstanceTypeSpec('azure:mem1_ssd1_x8', 15.7, 128, 8),
+              InstanceTypeSpec('azure:mem1_ssd1_x16', 31.4, 256, 16),
+
+              InstanceTypeSpec('azure:mem2_ssd1_x1', 3.5, 128, 1),
+              InstanceTypeSpec('azure:mem2_ssd1_x2', 7.0, 128, 2),
+              InstanceTypeSpec('azure:mem2_ssd1_x4', 14.0, 128, 4),
+              InstanceTypeSpec('azure:mem2_ssd1_x8', 28.0, 256, 8),
+              InstanceTypeSpec('azure:mem2_ssd1_x16', 56.0, 512, 16),
+
+              InstanceTypeSpec('azure:mem3_ssd1_x2', 14.0, 128, 2),
+              InstanceTypeSpec('azure:mem3_ssd1_x4', 28.0, 128, 4),
+              InstanceTypeSpec('azure:mem3_ssd1_x8', 56.0, 256, 8),
+              InstanceTypeSpec('azure:mem3_ssd1_x16', 112.0, 512, 16),
+              InstanceTypeSpec('azure:mem3_ssd1_x20', 140.0, 640, 20),
+
+              InstanceTypeSpec('azure:mem4_ssd1_x2', 28.0, 128, 2),
+              InstanceTypeSpec('azure:mem4_ssd1_x4', 56.0, 128, 4),
+              InstanceTypeSpec('azure:mem4_ssd1_x8', 112.0, 256, 8),
+              InstanceTypeSpec('azure:mem4_ssd1_x16', 224, 512, 16),
+              InstanceTypeSpec('azure:mem4_ssd1_x32', 448, 1024, 32)):
+        azure_preferred_instance_types[i.Name] = i
+    instance_types = OrderedDict()
+    instance_types.update(aws_preferred_instance_types)
+    instance_types.update(azure_preferred_instance_types)
+
+    for i in (InstanceTypeSpec('mem1_hdd2_x8', 7.0, 1680, 8),
+              InstanceTypeSpec('mem1_hdd2_x32', 60.5, 3360, 32),
+
+              InstanceTypeSpec('mem2_hdd2_x1', 3.8, 410, 1),
+              InstanceTypeSpec('mem2_hdd2_x2', 7.5, 840, 2),
+              InstanceTypeSpec('mem2_hdd2_x4', 15.0, 1680, 4),
 
               InstanceTypeSpec('mem3_hdd2_x2', 17.1, 420, 2),
               InstanceTypeSpec('mem3_hdd2_x4', 34.2, 850, 4),
               InstanceTypeSpec('mem3_hdd2_x8', 68.4, 1680, 8)):
         instance_types[i.Name] = i
+    default_instance_type = aws_preferred_instance_types['mem1_ssd1_x4']
     instance_type_names = instance_types.keys()
 
     def complete(self, text, state):
