@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2013-2015 DNAnexus, Inc.
+# Copyright (C) 2013-2016 DNAnexus, Inc.
 #
 # This file is part of dx-toolkit (DNAnexus platform client libraries).
 #
@@ -19,7 +19,7 @@
 
 from __future__ import print_function, unicode_literals, division, absolute_import
 
-import os, sys, unittest, json, tempfile, subprocess
+import os, sys, unittest, json, tempfile, subprocess, re
 import pexpect
 import pipes
 
@@ -28,6 +28,7 @@ import dxpy_testutil as testutil
 
 import dxpy
 from dxpy.scripts import dx_build_app
+from dxpy.utils.completer import InstanceTypesCompleter
 
 
 def run(command, **kwargs):
@@ -37,14 +38,14 @@ def run(command, **kwargs):
     return output
 
 
-supported_languages = ['Python', 'C++', 'bash']
+supported_languages = ['Python', 'bash']
 
-def run_dx_app_wizard():
+def run_dx_app_wizard(instance_type=None):
     old_cwd = os.getcwd()
     tempdir = tempfile.mkdtemp(prefix='Программа')
     os.chdir(tempdir)
     try:
-        wizard = pexpect.spawn("dx-app-wizard")
+        wizard = pexpect.spawn("dx-app-wizard --template parallelized")
         wizard.logfile = sys.stdout
         wizard.setwinsize(20, 90)
         wizard.expect("App Name:")
@@ -56,14 +57,8 @@ def run_dx_app_wizard():
         wizard.sendline("Заголовок")
         wizard.expect("Summary")
         wizard.sendline("Конспект")
-        wizard.expect("Description")
-        wizard.sendline("Описание")
         wizard.expect("Version")
         wizard.sendline("1.2.3")
-        wizard.expect("Choose a category")
-        wizard.sendline("Assembly")
-        wizard.expect("Choose a category")
-        wizard.sendline()
         wizard.expect("1st input name")
         wizard.sendline("in1")
         wizard.expect("Label")
@@ -84,19 +79,29 @@ def run_dx_app_wizard():
         wizard.sendline("int")
         wizard.expect("2nd output name")
         wizard.sendline()
+        wizard.expect("Timeout policy")
+        wizard.sendline("31d")
+        wizard.expect("Error: max allowed timeout is 30 days")
+        wizard.sendline("ЄЯTЪЦGЇCЄкЇ")
+        wizard.expect("Error: enter an int with a single-letter suffix")
+        wizard.expect("Timeout policy")
+        wizard.sendline("24h")
         wizard.expect("Programming language")
         wizard.sendline("АЛГОЛ")
         wizard.expect("Error: unrecognized response")
         wizard.sendline("Python")
-        wizard.expect("Execution pattern")
-        wizard.sendline("параллельно")
-        wizard.expect("Error: unrecognized response")
-        wizard.expect("Execution pattern")
-        wizard.sendline("parallelized")
         wizard.expect("Will this app need access to the Internet?")
         wizard.sendline("y")
         wizard.expect("Will this app need access to the parent project?")
         wizard.sendline("y")
+        wizard.expect("Choose an instance type for your app")
+        wizard.sendline("t1.микро")
+        wizard.expect("Error: unrecognized response, expected one of")
+        wizard.expect("Choose an instance type for your app")
+        if instance_type is not None:
+            wizard.sendline(instance_type)
+        else:
+            wizard.sendline()
         wizard.expect("App directory created")
         wizard.close()
 
@@ -139,12 +144,14 @@ def create_app_dir_with_dxapp_json(dxapp_json, language):
         wizard.sendline()
         wizard.expect("Version")
         wizard.sendline()
-        wizard.expect("Execution pattern")
+        wizard.expect("Timeout policy")
         wizard.sendline()
         wizard.expect("Will this app need access to the Internet?")
         wizard.sendline()
         wizard.expect("Will this app need access to the parent project?")
         wizard.sendline()
+        wizard.expect("Choose an instance type for your app")
+        wizard.sendline()
         wizard.expect("App directory created")
         wizard.close()
 
@@ -154,10 +161,24 @@ def create_app_dir_with_dxapp_json(dxapp_json, language):
         os.chdir(old_cwd)
 
 class TestDXAppWizardAndRunAppLocally(DXTestCase):
+    def test_invalid_arguments(self):
+        with self.assertRaises(testutil.DXCalledProcessError):
+            check_output(['dx-app-wizard', '--template=par'])
+
     def test_dx_app_wizard(self):
         appdir = run_dx_app_wizard()
         dxapp_json = json.load(open(os.path.join(appdir, 'dxapp.json')))
-        self.assertEqual(dxapp_json.get('authorizedUsers'), [])
+        self.assertEqual(dxapp_json['regionalOptions']['aws:us-east-1']['systemRequirements']['*']['instanceType'],
+                         InstanceTypesCompleter.default_instance_type.Name)
+        self.assertEqual(dxapp_json['runSpec']['distribution'], 'Ubuntu')
+        self.assertEqual(dxapp_json['runSpec']['release'], '14.04')
+        self.assertEqual(dxapp_json['runSpec']['timeoutPolicy']['*']['hours'], 24)
+
+    def test_dx_app_wizard_with_azure_instance_type(self):
+        appdir = run_dx_app_wizard("azure:mem1_ssd1_x2")
+        dxapp_json = json.load(open(os.path.join(appdir, 'dxapp.json')))
+        self.assertEqual(dxapp_json['regionalOptions']['azure:westus']['systemRequirements']['*']['instanceType'],
+                         "azure:mem1_ssd1_x2")
 
     def test_dx_run_app_locally_interactively(self):
         appdir = create_app_dir()
@@ -193,6 +214,17 @@ class TestDXAppWizardAndRunAppLocally(DXTestCase):
         result = remote_job.describe()
         self.assertEqual(result["output"]["out1"], 140)
 
+    def test_dx_build_app_locally_using_app_builder(self):
+        appdir = create_app_dir()
+        print("Setting current project to", self.project)
+        dxpy.WORKSPACE_ID = self.project
+        dxpy.PROJECT_CONTEXT_ID = self.project
+        bundled_resources = dxpy.app_builder.upload_resources(appdir)
+        applet_id, _ignored_applet_spec = dxpy.app_builder.upload_applet(appdir, bundled_resources, overwrite=True, dx_toolkit_autodep=False)
+        app_obj = dxpy.DXApplet(applet_id)
+        self.assertEqual(app_obj.describe()['id'], app_obj.get_id())
+
+
     def test_file_download(self):
         '''
         This test assumes a well-formed input spec and tests that the
@@ -296,14 +328,12 @@ class TestDXAppWizardAndRunAppLocally(DXTestCase):
         dxpy.PROJECT_CONTEXT_ID = self.project
 
         # Make some data objects for input
-        dxapplet = dxpy.api.applet_new({"project": dxpy.WORKSPACE_ID,
-                                        "name": "anapplet",
-                                        "dxapi": "1.0.0",
-                                        "runSpec": {"code": "", "interpreter": "bash"}})['id']
-        dxfile = dxpy.upload_string("foo", name="afile")
-        dxgtable = dxpy.new_dxgtable(columns=[{"name": "int_col", "type": "int"}], name="agtable")
-        dxgtable.add_rows([[3], [0]])
-        dxgtable.close(block=True)
+        dxpy.api.applet_new({"project": dxpy.WORKSPACE_ID,
+                             "name": "anapplet",
+                             "dxapi": "1.0.0",
+                             "runSpec": {"code": "", "interpreter": "bash",
+                                         "distribution": "Ubuntu", "release": "14.04"}})['id']
+        dxpy.upload_string("foo", name="afile")
         dxrecord = dxpy.new_dxrecord(name="arecord")
         dxrecord.close()
 
@@ -318,9 +348,9 @@ class TestDXAppWizardAndRunAppLocally(DXTestCase):
             "outputSpec": []
         }
 
-        classes = ['applet', 'record', 'file', 'gtable',
+        classes = ['applet', 'record', 'file',
                    'boolean', 'int', 'float', 'string', 'hash',
-                   'array:applet', 'array:record', 'array:file', 'array:gtable',
+                   'array:applet', 'array:record', 'array:file',
                    'array:boolean', 'array:int', 'array:float', 'array:string']
 
         for classname in classes:
@@ -343,10 +373,9 @@ class TestDXAppWizardAndRunAppLocally(DXTestCase):
                         '-irequired_array_record=arecord',
                         '-irequired_file=afile',
                         '-irequired_array_file=afile',
-                        '-irequired_gtable=agtable',
-                        '-irequired_array_gtable=agtable',
                         '-irequired_boolean=true',
                         '-irequired_array_boolean=true',
+                        '-irequired_array_boolean=false',
                         '-irequired_int=32',
                         '-irequired_array_int=42',
                         '-irequired_float=3.4',
@@ -359,6 +388,8 @@ class TestDXAppWizardAndRunAppLocally(DXTestCase):
             # Test with bare-minimum of inputs
             output = subprocess.check_output(['dx-run-app-locally', appdir] + cmdline_args)
             print(output)
+            # Verify array is printed total 3 times once in each input, logs, and final output
+            self.assertEquals(len(re.findall("required_array_boolean = \[ true, false ]", output)), 3)
             self.assertIn("App finished successfully", output)
 
             # See PTFM-13697 for CentOS 5 details
@@ -371,14 +402,8 @@ class TestDXAppWizardAndRunAppLocally(DXTestCase):
     @unittest.skipUnless(testutil.TEST_ENV, 'skipping test that would clobber your local environment')
     def test_dx_run_app_locally_without_auth(self):
         temp_file_path = tempfile.mkdtemp()
-        app_spec = {
-            "name": "test_run_locally_without_auth",
-            "dxapi": "1.0.0",
-            "runSpec": {"file": "code.py", "interpreter": "python2.7"},
-            "inputSpec": [{"name": "foo", "class": "file"}],
-            "outputSpec": [],
-            "version": "1.0.0"
-            }
+        app_spec = dict(testutil.DXTestCaseBuildApps.base_app_spec, name="test_run_locally_without_auth",
+                        inputSpec = [{"name": "foo", "class": "file"}])
         app_dir_path = os.path.join(temp_file_path, app_spec['name'])
         os.mkdir(app_dir_path)
         with open(os.path.join(app_dir_path, 'dxapp.json'), 'w') as manifest:
@@ -391,14 +416,10 @@ class TestDXAppWizardAndRunAppLocally(DXTestCase):
 
     def test_dx_run_app_locally_invalid_interpreter(self):
         temp_file_path = tempfile.mkdtemp()
-        app_spec = {
-            "name": "test_run_locally_invalid_interpreter",
-            "dxapi": "1.0.0",
-            "runSpec": {"file": "code.py", "interpreter": "python"},
-            "inputSpec": [],
-            "outputSpec": [],
-            "version": "1.0.0"
-            }
+        app_spec = dict(testutil.DXTestCaseBuildApps.base_app_spec,
+                        name="test_run_locally_invalid_interpreter",
+                        runSpec = {"file": "code.py", "interpreter": "python",
+                                    "distribution": "Ubuntu", "release": "14.04"})
         app_dir_path = os.path.join(temp_file_path, app_spec['name'])
         os.mkdir(app_dir_path)
         with open(os.path.join(app_dir_path, 'dxapp.json'), 'w') as manifest:
