@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2013-2015 DNAnexus, Inc.
+# Copyright (C) 2013-2016 DNAnexus, Inc.
 #
 # This file is part of dx-toolkit (DNAnexus platform client libraries).
 #
@@ -22,31 +22,15 @@ from __future__ import print_function, unicode_literals, division, absolute_impo
 import unittest, time, json, re, os
 import dateutil.parser
 import dxpy
-from dxpy import AppError, AppInternalError, DXFile, DXRecord
-from dxpy.utils import (describe, exec_utils, genomic_utils, response_iterator, get_futures_threadpool, DXJSONEncoder,
-                        normalize_timedelta, normalize_time_input, config)
+from dxpy import AppError, AppInternalError, DXError, DXFile, DXRecord
+from dxpy.utils import (exec_utils, genomic_utils, response_iterator, get_futures_threadpool, DXJSONEncoder,
+                        normalize_timedelta, normalize_time_input, config, Nonce)
 from dxpy.utils.exec_utils import DXExecDependencyInstaller
+from dxpy.utils.pretty_print import flatten_json_array
 from dxpy.compat import USING_PYTHON2
 
 # TODO: unit tests for dxpy.utils.get_field_from_jbor, get_job_from_jbor, is_job_ref
 
-class TestDescribe(unittest.TestCase):
-    def test_is_job_ref(self):
-        # Positive results
-        jobref = {"job": "job-B55ZF5kZKQGz1Xxyb5FQ0003", "field": "number"}
-        self.assertTrue(describe.is_job_ref(jobref))
-        jobref = {"$dnanexus_link": jobref}
-        self.assertTrue(describe.is_job_ref(jobref))
-
-        # Negative results
-        jobref = {"job": "job-B55ZF5kZKQGz1Xxyb5FQ0003", "field": "number", "other": "field"}
-        self.assertFalse(describe.is_job_ref(jobref))
-        jobref = {"job": "job-B55ZF5kZKQGz1Xxyb5FQ0003", "field": 32}
-        self.assertFalse(describe.is_job_ref(jobref))
-        jobref = {"$dnanexus_link": jobref}
-        self.assertFalse(describe.is_job_ref(jobref))
-        jobref = {"$dnanexus_link": "job-B55ZF5kZKQGz1Xxyb5FQ0003"}
-        self.assertFalse(describe.is_job_ref(jobref))
 
 class TestErrorSanitizing(unittest.TestCase):
     def test_error_sanitizing(self):
@@ -99,6 +83,32 @@ class TestResponseIterator(unittest.TestCase):
         for i, res in enumerate(response_iterator(tasks(), get_futures_threadpool(5), max_active_tasks=6)):
             self.assertEqual(i, res)
 
+    def test_first_chunk_sequentially(self):
+        pending_tasks = set()
+        complete_tasks = set()
+
+        def task(i):
+            pending_tasks.add(i)
+            if i >= 1:
+                self.assertNotIn(0, pending_tasks, "Task 0 was pending during task %d" % i)
+            time.sleep(1)
+            complete_tasks.add(i)
+            pending_tasks.remove(i)
+
+        def tasks():
+            for i in range(8):
+                yield task, [i], {}
+
+        #
+        # |---0---|
+        #           |---1---|
+        #           |---2---|
+        #           |---3---|
+        #
+        # Assert that 0 doesn't overlap with 1, 2, or 3
+        list(response_iterator(tasks(), get_futures_threadpool(8), do_first_task_sequentially=True))
+
+
 class TestDXUtils(unittest.TestCase):
     def test_dxjsonencoder(self):
         f = DXFile("file-" + "x"*24, project="project-" + "y"*24)
@@ -120,61 +130,85 @@ class TestEDI(DXExecDependencyInstaller):
         self.message_log.append(message)
 
 class TestDXExecDependsUtils(unittest.TestCase):
-    def test_dx_execdepends_installer(self):
-        def get_edi(run_spec, job_desc=None):
-            return TestEDI(executable_desc={"runSpec": run_spec}, job_desc=job_desc if job_desc else {})
+    if USING_PYTHON2:
+        assertRegex = unittest.TestCase.assertRegexpMatches
+        assertNotRegex = unittest.TestCase.assertNotRegexpMatches
 
+    def get_edi(self, run_spec, job_desc=None):
+        return TestEDI(executable_desc={"runSpec": run_spec}, job_desc=job_desc if job_desc else {})
+
+    def test_install_bundled_dependencies(self):
+        bundled_depends_by_region = {
+            "aws:us-east-1": [
+                {"name": "asset.east", "id": {"$dnanexus_link": "file-asseteast"}}
+            ],
+
+            "azure:westus": [
+                {"name": "asset.west", "id": {"$dnanexus_link": "file-assetwest"}}
+            ]
+        }
+        edi = self.get_edi({"bundledDependsByRegion": bundled_depends_by_region},
+                           job_desc={"region": "aws:us-east-1"})
+        with self.assertRaisesRegexp(DXError, 'file-asseteast'):
+            # Asserts that we attempted to download the correct file.
+            edi.install()
+        edi = self.get_edi({"bundledDependsByRegion": bundled_depends_by_region},
+                           job_desc={"region": "azure:westus"})
+        with self.assertRaisesRegexp(DXError, 'file-assetwest'):
+            edi.install()
+        with self.assertRaisesRegexp(KeyError, 'aws:cn-north-1'):
+            self.get_edi({"bundledDependsByRegion": bundled_depends_by_region},
+                         job_desc={"region": "aws:cn-north-1"})
+
+    def test_dx_execdepends_installer(self):
         def assert_cmd_ran(edi, regexp):
-            self.assertRegexpMatches("\n".join(edi.command_log), regexp)
+            self.assertRegex("\n".join(edi.command_log), regexp)
 
         def assert_log_contains(edi, regexp):
-            self.assertRegexpMatches("\n".join(edi.message_log), regexp)
+            self.assertRegex("\n".join(edi.message_log), regexp)
 
         with self.assertRaisesRegexp(AppInternalError, 'Expected field "runSpec" to be present'):
             DXExecDependencyInstaller({}, {})
 
         with self.assertRaisesRegexp(AppInternalError, 'Expected field "name" to be present'):
-            get_edi({"dependencies": [{"foo": "bar"}]})
+            self.get_edi({"dependencies": [{"foo": "bar"}]})
 
-        edi = get_edi({"dependencies": [{"name": "foo", "package_manager": "cran", "version": "1.2.3"}]})
+        edi = self.get_edi({"dependencies": [{"name": "foo", "package_manager": "cran", "version": "1.2.3"}]})
         edi.install()
         assert_cmd_ran(edi, "R -e .+ install.packages.+devtools.+install_version.+foo.+version.+1.2.3")
 
         with self.assertRaisesRegexp(AppInternalError, 'does not have a "url" field'):
-            get_edi({"dependencies": [{"name": "foo", "package_manager": "git"}]})
+            self.get_edi({"dependencies": [{"name": "foo", "package_manager": "git"}]})
 
-        edi = get_edi({"execDepends": [{"name": "git"}], "dependencies": [{"name": "tmux"}]})
+        edi = self.get_edi({"execDepends": [{"name": "git"}], "dependencies": [{"name": "tmux"}]})
         edi.install()
         assert_cmd_ran(edi, "apt-get install --yes --no-install-recommends git tmux")
 
-        edi = get_edi({})
+        edi = self.get_edi({"execDepends": [], "bundledDepends": [], "dependencies": []})
         edi.install()
 
-        edi = get_edi({"execDepends": [], "bundledDepends": [], "dependencies": []})
-        edi.install()
-
-        edi = get_edi({"dependencies": [{"name": "pytz", "package_manager": "pip", "version": "2014.7"},
-                                        {"name": "certifi", "package_manager": "pip", "stages": ["main"]},
-                                        {"name": "tmux", "package_manager": "apt"},
-                                        {"name": "rake", "package_manager": "gem", "version": "10.3.2"},
-                                        {"name": "nokogiri", "package_manager": "gem", "stages": ["main"]},
-                                        {"name": "Module::Provision", "package_manager": "cpan", "version": "0.36.1"},
-                                        {"name": "LWP::MediaTypes", "package_manager": "cpan"},
-                                        {"name": "RJSONIO", "package_manager": "cran"},
-                                        {"name": "plyr", "package_manager": "cran", "version": "1.8.1"},
-                                        {"name": "ggplot2", "package_manager": "cran", "stages": ["main"],
-                                         "version": "1.0.1"},
-                                        {"name": "r1", "id": {"$dnanexus_link": "record-123"}},
-                                        {"name": "g1",
-                                         "package_manager": "git",
-                                         "url": "https://github.com/dnanexus/oauth2-demo"},
-                                        {"name": "g2",
-                                         "package_manager": "git",
-                                         "url": "https://github.com/dnanexus/bwa",
-                                         "tag": "production",
-                                         "destdir": "/tmp/ee-edi-test-bwa",
-                                         "buld_commands": "echo build bwa here",
-                                         "stages": ["main"]}]})
+        edi = self.get_edi({"dependencies": [{"name": "pytz", "package_manager": "pip", "version": "2014.7"},
+                                             {"name": "certifi", "package_manager": "pip", "stages": ["main"]},
+                                             {"name": "tmux", "package_manager": "apt"},
+                                             {"name": "rake", "package_manager": "gem", "version": "10.3.2"},
+                                             {"name": "nokogiri", "package_manager": "gem", "stages": ["main"]},
+                                             {"name": "Module::Provision", "package_manager": "cpan", "version": "0.36.1"},
+                                             {"name": "LWP::MediaTypes", "package_manager": "cpan"},
+                                             {"name": "RJSONIO", "package_manager": "cran"},
+                                             {"name": "plyr", "package_manager": "cran", "version": "1.8.1"},
+                                             {"name": "ggplot2", "package_manager": "cran", "stages": ["main"],
+                                              "version": "1.0.1"},
+                                             {"name": "r1", "id": {"$dnanexus_link": "record-123"}},
+                                             {"name": "g1",
+                                              "package_manager": "git",
+                                              "url": "https://github.com/dnanexus/oauth2-demo"},
+                                             {"name": "g2",
+                                              "package_manager": "git",
+                                              "url": "https://github.com/dnanexus/bwa",
+                                              "tag": "production",
+                                              "destdir": "/tmp/ee-edi-test-bwa",
+                                              "buld_commands": "echo build bwa here",
+                                              "stages": ["main"]}]})
         edi.install()
         assert_cmd_ran(edi, re.escape("pip install --upgrade pytz==2014.7 certifi"))
         assert_cmd_ran(edi, "apt-get install --yes --no-install-recommends tmux")
@@ -185,19 +219,43 @@ class TestDXExecDependsUtils(unittest.TestCase):
         assert_cmd_ran(edi, "cd /tmp/ee-edi-test-bwa && git clone https://github.com/dnanexus/bwa")
         assert_cmd_ran(edi, "git checkout production")
 
-        edi = get_edi({"execDepends": [{"name": "w00t", "stages": ["foo", "bar"]},
-                                       {"name": "f1", "id": {"$dnanexus_link": "file-123"}, "stages": ["xyzzt"]}]})
+        edi = self.get_edi({"execDepends": [{"name": "w00t", "stages": ["foo", "bar"]},
+                                            {"name": "f1", "id": {"$dnanexus_link": "file-123"}, "stages": ["xyzzt"]}]})
         edi.install()
-        self.assertNotRegexpMatches("\n".join(edi.command_log), "w00t")
+        self.assertNotRegex("\n".join(edi.command_log), "w00t")
         for name in "w00t", "f1":
             assert_log_contains(edi,
                                 "Skipping dependency {} because it is inactive in stage \(function\) main".format(name))
 
-        edi = get_edi({"execDepends": [{"name": "git", "stages": ["foo", "bar"]}]},
-                      job_desc = {"function": "foo"})
+        edi = self.get_edi({"execDepends": [{"name": "git", "stages": ["foo", "bar"]}]},
+                           job_desc={"function": "foo"})
         edi.install()
         assert_cmd_ran(edi, "apt-get install --yes --no-install-recommends git")
 
+        # Job describe dict must contain "region" if the run specification
+        # contains "bundledDependsByRegion".
+        bundled_depends_by_region = {
+            "aws:us-east-1": [
+                {"name": "asset.east", "id": {"$dnanexus_link": "file-asseteast"}}
+            ]
+        }
+        with self.assertRaisesRegexp(DXError, 'region.*job description'):
+            self.get_edi({"bundledDependsByRegion": bundled_depends_by_region})
+
+        bundled_depends_by_region = {
+            "aws:us-east-1": []
+        }
+        with self.assertRaisesRegexp(DXError, 'region.*job description'):
+            self.get_edi({"bundledDependsByRegion": bundled_depends_by_region})
+
+        # Job describe dict may specify or omit "region" if
+        # "bundledDependsByRegion" is not in run specification.
+        edi = self.get_edi({}, job_desc={"region": "azure:westus"})
+        edi.install()
+
+        edi = self.get_edi({}, job_desc={})
+        edi.install()
+
 
 class TestTimeUtils(unittest.TestCase):
     def test_normalize_timedelta(self):
@@ -277,7 +335,7 @@ class TestDXConfig(unittest.TestCase):
             self.assertIn("DXConfig object at", repr(c))
             c.update(c.defaults)
             self.assertEqual(len(c), len(list(c)))
-            del c[c.defaults.keys()[0]]
+            del c[list(c.defaults.keys())[0]]
             self.assertEqual(len(c), len(list(c)))
             dxpy.config["DX_PROJECT_CONTEXT_NAME"] = None
             self.assertEqual(os.environ["DX_PROJECT_CONTEXT_NAME"], "")
@@ -285,5 +343,60 @@ class TestDXConfig(unittest.TestCase):
             os.environ.update(environ_backup)
             dxpy.config.__init__(suppress_warning=True)
 
+class TestPrettyPrint(unittest.TestCase):
+    def test_flatten_json_array(self):
+        json_string = (
+            '{\n'
+            '  "arr": [\n'
+            '    "one",\n'
+            '    2,\n'
+            '    3.4,\n'
+            '    "\\"five\\""\n'
+            '  ],\n'
+            '  "foo": {\n'
+            '    "arr": [\n'
+            '      "six",\n'
+            '      7.8,\n'
+            '      9,\n'
+            '      "t\\"en"\n'
+            '    ]\n'
+            '  }\n'
+            '}'
+        )
+        flattened_json_string_ref = (
+            '{\n'
+            '  "arr": ["one", 2, 3.4, "\\"five\\""],\n'
+            '  "foo": {\n'
+            '    "arr": ["six", 7.8, 9, "t\\"en"]\n'
+            '  }\n'
+            '}'
+        )
+        flatten_json_array(json_string, "arr")
+        self.assertEqual(flattened_json_string_ref, flatten_json_array(json_string, "arr"))
+
+
+class TestNonceGeneration(unittest.TestCase):
+    def test_nonce_generator(self):
+        nonce_list = []
+        for i in range(0, 100):
+            nonce_list.append(str(Nonce()))
+
+        for nonce in nonce_list:
+            self.assertTrue(len(nonce) > 0)
+            self.assertTrue(len(nonce) <= 128)
+            self.assertEqual(nonce_list.count(nonce), 1)
+
+    def test_input_updater(self):
+        input_params = {"p1": "v1", "p2": "v2"}
+        updated_input = Nonce.update_nonce(input_params)
+        self.assertIn("nonce", updated_input)
+
+        nonce = str(Nonce())
+        input_params.update({"nonce": nonce})
+        updated_input = Nonce.update_nonce(input_params)
+        self.assertIn("nonce", updated_input)
+        self.assertEqual(nonce, updated_input["nonce"])
+
+
 if __name__ == '__main__':
     unittest.main()
